#pragma section REPRO
#include "memory_dump.h"

#include <string.h> // for memcpy

#include "../System/TimeManager/time_manager.h"
#include "../CmdTlm/packet_handler.h"
#include "../Library/endian_memcpy.h"

static MemoryDump memory_dump_;
const MemoryDump* const memory_dump = &memory_dump_;

static CTCP MEM_ctcp_; // f[^TCY傫̂staticm

static void MEM_init_(void);

static uint8_t MEM_get_next_adu_counter_(void);

static CCP_EXEC_STS MEM_dump_region_(uint8_t category,
                                     uint8_t num_dumps);

static MEM_ACK MEM_form_packet_(TCP* packet,
                                uint8_t category);

static TCP_SEQ_FLAG MEM_judge_seq_flag_(uint32_t rp,
                                        size_t len);

static void MEM_setup_header_(TCP* packet,
                              size_t data_len,
                              uint8_t category,
                              uint8_t adu_cnt,
                              TCP_SEQ_FLAG adu_seq_flag,
                              uint16_t adu_seq_cnt,
                              uint32_t adu_len);

static void MEM_send_packet_(const CTCP* packet,
                             uint8_t num_dumps);

AppInfo MEM_create_app(void)
{
  return create_app_info("mem", MEM_init_, NULL);
}

static void MEM_init_(void)
{
  memory_dump_.begin = 0;
  memory_dump_.end = 0;
  memory_dump_.adu_size = 0;
  memory_dump_.adu_seq = 0;
  memory_dump_.adu_counter = 0;
}

CCP_EXEC_STS Cmd_MEM_SET_REGION(const CTCP* packet)
{
  const uint8_t* param = CCP_get_param_head(packet);
  uint32_t begin, end, span;

  // p[^ǂݏo
  endian_memcpy(&begin, param, 4);
  endian_memcpy(&end,   param + 4, 4);

  if (begin > end)
  {
    // ̈̊JnƏȊ召֌Wt̏ꍇُ͈I
    return CCP_EXEC_ILLEGAL_CONTEXT;
  }

  span = end - begin;

  if (span > MEM_MAX_SPAN)
  {
    // w_vőʂ𒴂Ăꍇُ͈IB
    return CCP_EXEC_ILLEGAL_PARAMETER;
  }

  memory_dump_.begin = begin;
  memory_dump_.end = end;
  memory_dump_.adu_size = span;
  memory_dump_.adu_seq = 0;

  // ̈ݒ1񖈂ɓƗADUJEgl蓖ĂB
  memory_dump_.adu_counter = MEM_get_next_adu_counter_();

  return CCP_EXEC_SUCCESS;
}

CCP_EXEC_STS Cmd_MEM_DUMP_REGION_SEQ(const CTCP* packet)
{
  const uint8_t* param = CCP_get_param_head(packet);
  uint8_t category, num_dumps;

  // p[^ǂݏo
  category = param[0];
  num_dumps = param[1];

  if (num_dumps >= 8)
  {
    // pPbg񐔂̏8ƂB
    // 32kbpsłDL8VCDU/sec1b̒ʐMʁB
    // 𒴂ꍇ͕R}h𑗐MđΉB
    return CCP_EXEC_ILLEGAL_PARAMETER;
  }

  return MEM_dump_region_(category, num_dumps);
}

CCP_EXEC_STS Cmd_MEM_DUMP_REGION_RND(const CTCP* packet)
{
  const uint8_t* param = CCP_get_param_head(packet);
  uint8_t category, num_dumps;
  uint16_t adu_seq;
  uint32_t rp;

  // p[^ǂݏo
  category = param[0];
  num_dumps = param[1];

  if (num_dumps >= 8)
  {
    // pPbg񐔂̏8ƂB
    // 32kbpsłDL8VCDU/sec1b̒ʐMʁB
    // 𒴂ꍇ͕R}h𑗐MđΉB
    return CCP_EXEC_ILLEGAL_PARAMETER;
  }

  endian_memcpy(&adu_seq, param + 2, 2);

  rp = memory_dump_.begin + (adu_seq * MEM_DUMP_WIDTH);

  if (rp < memory_dump_.end)
  {
    // w肳ꂽADU Sequence Counter̈Ȃl荞݁A
    // w肳ꂽʒu̓e_vB
    // uw肷 = _vvƔfĂ
    memory_dump_.adu_seq = adu_seq;
    return MEM_dump_region_(category, num_dumps);
  }
  else
  {
    // w肳ꂽADU Sequence Counter̈OłΈُI
    return CCP_EXEC_ILLEGAL_CONTEXT;
  }
}

CCP_EXEC_STS Cmd_MEM_DUMP_SINGLE(const CTCP* packet)
{
  const uint8_t* param = CCP_get_param_head(packet);
  uint8_t category, num_dumps;
  uint32_t start_addr;
  uint8_t* data;

  // p[^ǂݏo
  category = param[0];
  num_dumps = param[1];

  if (num_dumps >= 8)
  {
    // pPbg񐔂̏8ƂB
    // 32kbpsłDL8VCDU/sec1b̒ʐMʁB
    // 𒴂ꍇ͕R}h𑗐MđΉB
    return CCP_EXEC_ILLEGAL_PARAMETER;
  }

  endian_memcpy(&start_addr, param + 2, 4);

  // v: wAhXlُȏꍇׂ̏H
  // Segmentation FaultƂNH

  // _vpPbg̃wb_ݒ
  // Packet Sequence Counter for each ADU͒PpPbgȂ̂0ŒB
  // ADU`ɂȂ5Bytes_v̈𑝂₹邪A
  // Packet IDʂɂKv̂łЂƂ܂`킹ĎB
  MEM_setup_header_(&MEM_ctcp_,
                    MEM_DUMP_WIDTH,              // 1pPbg̍ő咷_v
                    category,                    // oJeS̓p[^Ŏw
                    MEM_get_next_adu_counter_(), // PADUȂ̂ADUJEgl𒼐ڊ蓖
                    TCP_SEQ_SINGLE,              // AhXw̏ꍇ͒PADUƂď
                    0x0000,                      // PADUȂ̂0x0000ɌŒ
                    MEM_DUMP_WIDTH);             // ADUDUMPƓ

  // _vf[^ݒ
  data = TCP_TLM_get_user_data_head(&MEM_ctcp_) + 5; // ADUɂwb_5Bytes
  memcpy(data, (const void*)start_addr, MEM_DUMP_WIDTH);

  // pPbg𑗏o
  MEM_send_packet_(&MEM_ctcp_, num_dumps);

  return CCP_EXEC_SUCCESS;
}

CCP_EXEC_STS Cmd_MEM_LOAD(const CTCP* packet)
{
  const uint8_t* param = CCP_get_param_head(packet);
  size_t param_len = CCP_get_param_len(packet);
  size_t data_len;
  uint32_t start_addr;

  // f[^ǂݏo
  // f[^ -> p[^S-JnAhXp[^
  data_len = param_len - 4;

  // ݃AhXǂݏo
  endian_memcpy(&start_addr, param, 4);

  // w肵JnAhXn܂̈Ƀf[^
  memcpy((void*)start_addr, &(param[4]), data_len);
  return CCP_EXEC_SUCCESS;
}

CCP_EXEC_STS Cmd_MEM_SET_DESTINATION(const CTCP* packet)
{
  const uint8_t* param = CCP_get_param_head(packet);
  uint32_t dest;

  endian_memcpy(&dest, param, 4);

  if ((dest >= memory_dump_.begin) && (dest < memory_dump_.end))
  {
    // AhẌɊ܂܂ꍇB
    // F߂ƏGɂȂ̂ŋ֎~B
    return CCP_EXEC_ILLEGAL_PARAMETER;
  }

  // AhXݒ肵ARP̈擪ɍ킹B
  memory_dump_.dest = dest;
  memory_dump_.rp = memory_dump_.begin;
  return CCP_EXEC_SUCCESS;
}

CCP_EXEC_STS Cmd_MEM_COPY_REGION_SEQ(const CTCP* packet)
{
  const uint8_t* param = CCP_get_param_head(packet);
  uint32_t copy_width, wp;

  if (memory_dump_.rp == memory_dump_.end)
  {
    // ɗ̈Ŝ̓ǂݏoĂꍇB
    // ͍s킸IB
    return CCP_EXEC_SUCCESS;
  }

  // p[^ǂݏoB
  endian_memcpy(&copy_width, param, 4);

  if ((memory_dump_.rp + copy_width) > memory_dump_.end)
  {
    // wRs[ŃRs[Ɨ̈OɏoĂ܂ꍇB
    // Rs[ʂ̈I[Ɉv悤Rs[NbvB
    copy_width = memory_dump_.end - memory_dump_.rp;
  }

  // WPvZBWP = 擪+ǂݏoς݃f[^ʁB
  wp = memory_dump_.dest + (memory_dump_.rp - memory_dump_.begin);

  // w肳ꂽRs[̈Rs[ARPXVB
  memcpy((uint8_t*)wp, (const uint8_t*)memory_dump_.rp, copy_width);
  memory_dump_.rp += copy_width;
  return CCP_EXEC_SUCCESS;
}

static uint8_t MEM_get_next_adu_counter_(void)
{
  // CNg̒lԂ̂ŏl0xffƂB
  static uint8_t adu_counter_ = 0xff;
  return ++adu_counter_;
}

static CCP_EXEC_STS MEM_dump_region_(uint8_t category,
                                     uint8_t num_dumps)
{
  MEM_ACK ack;

  // ݒlɂƂÂoׂpPbg\z
  ack = MEM_form_packet_(&MEM_ctcp_, category);

  switch (ack)
  {
  case MEM_SUCCESS:
    // pPbg𑗏oAADU Sequence Counter̒lXV
    MEM_send_packet_(&MEM_ctcp_, num_dumps);
    ++memory_dump_.adu_seq;
    return CCP_EXEC_SUCCESS;

  case MEM_NO_DATA:
    // łɑS̈_vς݂Ȃ牽I
    return CCP_EXEC_SUCCESS;

  default:
    // ȊÕG[͂Ȃ͂
    return CCP_EXEC_UNKNOWN;
  }
}

static MEM_ACK MEM_form_packet_(TCP* packet,
                                uint8_t category)
{
  uint32_t rp;
  size_t len;
  TCP_SEQ_FLAG seq_flag;
  uint8_t* data;

  rp = memory_dump_.begin + (memory_dump_.adu_seq * MEM_DUMP_WIDTH);

  if (rp >= memory_dump_.end)
  {
    // Read PointerI[ɒBĂΉI
    return MEM_NO_DATA;
  }

  // c_vvZ
  len = memory_dump_.end - rp;

  if (len > MEM_DUMP_WIDTH)
  {
    // 1pPbg̃_v𒴂ꍇ̓_vɉB
    len = MEM_DUMP_WIDTH;
  }

  // Packet Sequence Flag̎ʔ
  seq_flag = MEM_judge_seq_flag_(rp, len);

  // _vpPbg̃wb_ݒ
  MEM_setup_header_(&MEM_ctcp_,
                    len,
                    category,
                    memory_dump_.adu_counter,
                    seq_flag,
                    memory_dump_.adu_seq,
                    memory_dump_.adu_size);

  // _vf[^ݒ
  data = TCP_TLM_get_user_data_head(packet) + 5; // ADUɂwb_5Bytes
  memcpy(data, (const void*)rp, len);

  return MEM_SUCCESS;
}

static TCP_SEQ_FLAG MEM_judge_seq_flag_(uint32_t rp,
                                        size_t len)
{
  if (memory_dump_.adu_size < MEM_DUMP_WIDTH)
  {
    // ADUPpPbgɎ܂ꍇ
    return TCP_SEQ_SINGLE;
  }
  else
  {
    // ADUpPbgɂ܂ꍇ
    if (rp == memory_dump_.begin)
    {
      // Read Pointerǂݏö擪ɈvȂJnpPbg
      return TCP_SEQ_FIRST;
    }
    else if (len < MEM_DUMP_WIDTH)
    {
      // _vcʂ1pPbgȓɎ܂ȂIpPbg
      return TCP_SEQ_LAST;
    }
    else
    {
      // ȊȌꍇ͒ԃpPbg
      return TCP_SEQ_CONT;
    }
  }
}

static void MEM_setup_header_(TCP* packet,
                              size_t data_len,
                              uint8_t category,
                              uint8_t adu_cnt,
                              TCP_SEQ_FLAG adu_seq_flag,
                              uint16_t adu_seq_cnt,
                              uint32_t adu_len)
{
  uint8_t* data;

  // TCPacketwb_ݒ
  TCP_TLM_setup_primary_hdr(packet, TCP_APID_PROCYON_DUMP_TLM, (uint16_t)(data_len + 12));
  TCP_TLM_set_ti(packet, (uint32_t)( TMGR_get_master_total_cycle() ));
  TCP_TLM_set_category(packet, category); // p[^ɂw
  TCP_TLM_set_packet_id(packet, (uint8_t)MEM_TLM_ID);
  TCP_TLM_set_adu_cnt(packet, adu_cnt);
  TCP_TLM_set_adu_seq_flag(packet, adu_seq_flag);

  // ADUSecondary Headerǉݒ
  data = TCP_TLM_get_user_data_head(packet);
  data[0] = (uint8_t)(adu_seq_cnt >> 8);
  data[1] = (uint8_t)(adu_seq_cnt);
  data += 2;
  // ADU Lengthݒ 3BytesȂ̂Œ
  data[0] = (uint8_t)(adu_len >> 16);
  data[1] = (uint8_t)(adu_len >> 8);
  data[2] = (uint8_t)(adu_len);
}

static void MEM_send_packet_(const CTCP* packet,
                             uint8_t num_dumps)
{
  int i;

  // TLMo
  for (i = 0; i < num_dumps; ++i)
  {
    // num_dumps̉񐔂pPbgz֓n
    PH_analyze_packet(packet);
  }
}

#pragma section
