#pragma section REPRO
/**
 * @file   driver_super.c
 * @brief  eZTEAN`G[^Ƃ̃C^[tF[XhCoQ̃X[p[NX
 *
 *         DriverSuper Class 
 *         eZTEAN`G[^Ƃ̃C^[tF[XC
 *         CR}hsCegNGXgCegMCeg͂ȂǂsChCoQ̃X[p[NXłD
 *         X̋@̃C^[tF[XhCoɌpĎgp܂D
 */

#include "driver_super.h"
#include "../../Library/print.h"
#include <string.h>     // for memsetȂǂmemn
#include <stddef.h>     // for NULL

// #define DS_DEBUG                       // K؂ȂƂɃRgAEg
// #define DS_DEBUG_SHOW_REC_DATA         // K؂ȂƂɃRgAEg

static DS_ERR_CODE DS_send_cmd_(DriverSuper* p_super, uint8_t stream);
static int      DS_tx_(DriverSuper* p_super, uint8_t stream);
static int      DS_rx_(DriverSuper* p_super);
static void     DS_analyze_rx_buffer_(DriverSuper* p_super,
                                      uint8_t stream,
                                      uint16_t rec_data_len);
static uint16_t DS_analyze_rx_buffer_prepare_buffer_(DriverSuper* p_super,
                                                     uint8_t stream,
                                                     uint8_t* rx_buffer,
                                                     uint16_t rec_data_len);
static uint16_t DS_analyze_rx_buffer_pickup_(DS_StreamConfig* p_stream_config,
                                             uint8_t* rx_buffer,
                                             uint16_t rec_data_len);
static void     DS_analyze_rx_buffer_carry_over_buffer_(DS_StreamConfig* p_stream_config,
                                                        uint8_t* rx_buffer,
                                                        uint16_t total_processed_data_len,
                                                        uint16_t rec_data_len);
static uint16_t DS_analyze_rx_buffer_fixed_pickup_(DS_StreamConfig* p_stream_config,
                                                   uint8_t* rx_buffer,
                                                   uint16_t total_processed_data_len,
                                                   uint16_t rec_data_len);
static uint16_t DS_analyze_rx_buffer_variable_pickup_with_rx_frame_size_(DS_StreamConfig* p_stream_config,
                                                                         uint8_t* rx_buffer,
                                                                         uint16_t total_processed_data_len,
                                                                         uint16_t rec_data_len);
static uint16_t DS_analyze_rx_buffer_variable_pickup_with_footer_(DS_StreamConfig* p_stream_config,
                                                                  uint8_t* rx_buffer,
                                                                  uint16_t total_processed_data_len,
                                                                  uint16_t rec_data_len);
static uint16_t DS_analyze_rx_buffer_finding_header_(DS_StreamConfig* p_stream_config,
                                                     uint8_t* rx_buffer,
                                                     uint16_t total_processed_data_len,
                                                     uint16_t rec_data_len);
static uint16_t DS_analyze_rx_buffer_receiving_header_(DS_StreamConfig* p_stream_config,
                                                       uint8_t* rx_buffer,
                                                       uint16_t total_processed_data_len);
static uint16_t DS_analyze_rx_buffer_receiving_footer_(DS_StreamConfig* p_stream_config,
                                                       uint8_t* rx_buffer,
                                                       uint16_t total_processed_data_len,
                                                       uint16_t rx_frame_size);
static uint32_t DS_analyze_rx_buffer_get_framelength_(DS_StreamConfig* p_stream_config);

static DS_ERR_CODE DS_reset_stream_config_(DS_StreamConfig* p_stream_config);
static DS_ERR_CODE DS_validate_stream_config_(DS_StreamConfig* p_stream_config);

// _~[֐
// EQUƊ֐|C^̏lNULLɂĂ߂ɂʂۂŎ̂̂
static DS_ERR_CODE DS_load_init_setting_dummy_(DriverSuper* p_super);
static DS_ERR_CODE DS_data_analyzer_dummy_(DS_StreamConfig* p_stream_config, void* p_driver);


// ###### DriverSuper{֐ ######

DS_ERR_CODE DS_init(DriverSuper* p_super, void* if_config, DS_ERR_CODE (*load_init_setting)(DriverSuper* p_super))
{
  if (DS_reset(p_super) != DS_ERR_CODE_OK) return DS_ERR_CODE_ERR;

  p_super->if_config = if_config;

  p_super->config.load_init_setting = load_init_setting;
  if (p_super->config.load_init_setting(p_super) != DS_ERR_CODE_OK) return DS_ERR_CODE_ERR;

  if (DS_validate_config(p_super) != DS_ERR_CODE_OK) return DS_ERR_CODE_ERR;

  // IF̏
  // UVvIF_init̃G[R[h͖i@łŃG[oꍇ̓R[ĥŁDKvΏDj
  if ( (*IF_init[p_super->interface])(p_super->if_config) != 0 ) return DS_ERR_CODE_ERR;

  return DS_ERR_CODE_OK;
}


DS_ERR_CODE DS_reset(DriverSuper* p_super)
{
  uint8_t stream;

  p_super->interface = IF_LIST_MAX; // FIXME: (*IF_init[p_super->interface])(p_super->if_config) ̗lȎĝŃZOtHN\
  p_super->if_config = NULL;        // FIXME: NULL|C^͂̊֐ResetP̂Ŏgƃ}Y

  memset(p_super->config.rx_buffer_, 0x00, sizeof(p_super->config.rx_buffer_));

  p_super->config.load_init_setting = DS_load_init_setting_dummy_;

  p_super->config.rec_status_.ret_from_if_rx       = 0;
  p_super->config.rec_status_.rx_disruption_status = DS_RX_DISRUPTION_STATUS_OK;

  p_super->config.rx_count_      = 0;
  p_super->config.rx_call_count_ = 0;
  p_super->config.rx_time_       = TMGR_get_master_clock();

  p_super->config.should_monitor_for_rx_disruption_ = 0;
  p_super->config.time_threshold_for_rx_disruption_ = 60 * 1000;      // ̒l͂悭l邱

  for (stream = 0; stream < DS_STREAM_MAX; ++stream)
  {
    DS_ERR_CODE ret = DS_reset_stream_config_(&p_super->stream_config[stream]);
    if (ret != DS_ERR_CODE_OK) return ret;
  }
  return DS_ERR_CODE_OK;
}


DS_ERR_CODE DS_validate_config(DriverSuper* p_super)
{
  uint8_t stream;

  if (p_super->interface < 0 || p_super->interface >= IF_LIST_MAX) return DS_ERR_CODE_ERR;
  if (p_super->if_config == NULL) return DS_ERR_CODE_ERR;

  for (stream = 0; stream < DS_STREAM_MAX; ++stream)
  {
    DS_ERR_CODE ret = DS_validate_stream_config_(&p_super->stream_config[stream]);
    if (ret != DS_ERR_CODE_OK) return ret;
  }

  return DS_ERR_CODE_OK;
}

DS_ERR_CODE DS_clear_rx_buffer(DriverSuper* p_super)
{
  uint8_t stream;

  // ȉCe buffer  memsetŔÖ0NAĂC
  //  carry_over_buffer_size_ ɂ̂ŁCӖ͂ȂD
  memset(p_super->config.rx_buffer_, 0x00, sizeof(p_super->config.rx_buffer_));

  for (stream = 0; stream < DS_STREAM_MAX; ++stream)
  {
    p_super->stream_config[stream].rx_frame_rec_len_ = 0;
    p_super->stream_config[stream].is_rx_buffer_carry_over_ = 0;
    p_super->stream_config[stream].carry_over_buffer_size_  = 0;

    memset(p_super->stream_config[stream].rx_frame_,
           0x00,
           sizeof(p_super->stream_config[stream].rx_frame_));
    memset(p_super->stream_config[stream].rx_buffer_for_carry_over_,
           0x00,
           sizeof(p_super->stream_config[stream].rx_buffer_for_carry_over_));
  }

  return DS_ERR_CODE_OK;
}


DS_ERR_CODE DS_receive(DriverSuper* p_super)
{
  uint8_t  stream;
  uint16_t rec_data_len;
  int      ret_rx;

  p_super->config.rx_call_count_++;

  // eDriverŕIɐڑĂ wire ͂P{Ȃ̂ŁCŎMD
  // i stream ł́C̎Mrbgɑ΂āC̃t[ނɑ΂āCt[TCm菈𑖂炷D
  ret_rx = DS_rx_(p_super);
  p_super->config.rec_status_.ret_from_if_rx = ret_rx;

  if (ret_rx > 0)
  {
    // Ȃɂ̎Mf[^
    p_super->config.rx_count_++;
    p_super->config.rx_time_ = TMGR_get_master_clock();
  }

  // Mr┻
  // eȂǂŌƂɃmCYɂȂ̂ŁC肵ȂƂ OK ɂĂ
  p_super->config.rec_status_.rx_disruption_status = DS_RX_DISRUPTION_STATUS_OK;
  if (p_super->config.should_monitor_for_rx_disruption_)
  {
    ObcTime now = TMGR_get_master_clock();
    uint32_t last_rx_ago = OBCT_diff_in_msec(&p_super->config.rx_time_, &now);

    if (last_rx_ago > p_super->config.time_threshold_for_rx_disruption_)
    {
      p_super->config.rec_status_.rx_disruption_status = DS_RX_DISRUPTION_STATUS_LOST;
    }
  }

  for (stream = 0; stream < DS_STREAM_MAX; ++stream)
  {
    DS_StreamConfig* p_stream_config = &(p_super->stream_config[stream]);
    if (!p_stream_config->is_enabled_)
    {
      p_stream_config->rec_status_.status_code = DS_STREAM_REC_STATUS_DISABLE;
      continue;
    }

    // setter  validation ƁCȂǂŉx邱ƂC
    //  validation ͑łグƂނn㎎ɗLpȂ̂ŁCɒu
    if (p_stream_config->is_validation_needed_for_rec_)
    {
      DS_ERR_CODE ret = DS_validate_stream_config_(p_stream_config);
      if (ret != DS_ERR_CODE_OK)
      {
        p_stream_config->rec_status_.status_code = DS_STREAM_REC_STATUS_VALIDATE_ERR;
        continue;
      }
    }

    if (ret_rx < 0)     // ̋ʏ for ̒ɂ́C኱LDDD
    {
      // RXs
      p_stream_config->rec_status_.status_code = DS_STREAM_REC_STATUS_RX_ERR;
      // [TODO]  DriverSuper ƂĂ̋ʂ̃Am}s邩͗vc_
      //        ʑΉȂ̂ŁC Driver ɎCȂ́C IF Ɏ̂RH
      continue;
    }
    else if (ret_rx == 0)
    {
      // Mf[^Ȃ
      // Jzf[^Ώ
      if (p_stream_config->is_rx_buffer_carry_over_)
      {
        // JẑŁC continue ֎
      }
      else
      {
        // rec_status_.status_code {͍XV
        // FIXED ̏ꍇ̓Zbg
        if (p_stream_config->rec_status_.status_code == DS_STREAM_REC_STATUS_FIXED_FRAME)
        {
          p_stream_config->rec_status_.status_code = DS_STREAM_REC_STATUS_FINDING_HEADER;
        }
        continue;
      }
    }
    else
    {
      // Ȃ
    }

    // ܂łMf[^

    rec_data_len = (uint16_t)ret_rx;      // ܂łΔ񕉐
    DS_analyze_rx_buffer_(p_super, stream, rec_data_len);

    // t[m菈
    if (p_stream_config->rec_status_.status_code == DS_STREAM_REC_STATUS_FIXED_FRAME)
    {
      p_stream_config->rx_frame_fix_count_++;
      p_stream_config->rx_frame_fix_time_ = TMGR_get_master_clock();
      p_stream_config->req_tlm_cmd_tx_count_after_last_tx_ = 0;
    }
  }

  // stream̃er┻iet[mr┻j
  for (stream = 0; stream < DS_STREAM_MAX; ++stream)
  {
    DS_StreamConfig* p_stream_config = &(p_super->stream_config[stream]);
    ObcTime now;
    uint32_t last_tlm_fix_ago;

    // eȂǂŌƂɃmCYɂȂ̂ŁC肵ȂƂ OK ɂĂ
    p_stream_config->rec_status_.tlm_disruption_status = DS_STREAM_TLM_DISRUPTION_STATUS_OK;

    if (!p_stream_config->is_enabled_ || !p_stream_config->should_monitor_for_tlm_disruption_)
    {
      continue;
    }

    now = TMGR_get_master_clock();
    last_tlm_fix_ago = OBCT_diff_in_msec(&p_stream_config->rx_frame_fix_time_, &now);

    if (last_tlm_fix_ago > p_stream_config->time_threshold_for_tlm_disruption_)
    {
      p_stream_config->rec_status_.tlm_disruption_status = DS_STREAM_TLM_DISRUPTION_STATUS_LOST;
    }
  }

  return (ret_rx < 0) ? DS_ERR_CODE_ERR : DS_ERR_CODE_OK;
}


DS_ERR_CODE DS_analyze_rec_data(DriverSuper* p_super, uint8_t stream, void* p_driver)
{
  DS_StreamConfig* p_stream_config = &(p_super->stream_config[stream]);

  p_stream_config->ret_from_data_analyzer_ = p_stream_config->data_analyzer_(p_stream_config, p_driver);

  return p_stream_config->ret_from_data_analyzer_;
}


DS_ERR_CODE DS_send_general_cmd(DriverSuper* p_super, uint8_t stream)
{
  DS_StreamConfig* p_stream_config = &(p_super->stream_config[stream]);

  if (!p_stream_config->is_enabled_)
  {
    p_stream_config->send_status_.status_code = DS_STREAM_SEND_STATUS_DISABLE;
    return DS_ERR_CODE_OK;
  }

  p_stream_config->general_cmd_tx_count_++;
  p_stream_config->general_cmd_tx_time_ = TMGR_get_master_clock();

#ifdef DS_DEBUG
  Printf("DS: send_general_cmd\n");
#endif

  return DS_send_cmd_(p_super, stream);
}


DS_ERR_CODE DS_send_req_tlm_cmd(DriverSuper* p_super, uint8_t stream)
{
  DS_StreamConfig* p_stream_config = &(p_super->stream_config[stream]);

  if (!p_stream_config->is_enabled_)
  {
    p_stream_config->send_status_.status_code = DS_STREAM_SEND_STATUS_DISABLE;
    return DS_ERR_CODE_OK;
  }

  p_stream_config->req_tlm_cmd_tx_count_++;
  p_stream_config->req_tlm_cmd_tx_count_after_last_tx_++;
  p_stream_config->req_tlm_cmd_tx_time_ = TMGR_get_master_clock();

#ifdef DS_DEBUG
  Printf("DS: send_req_tlm_cmd\n");
#endif

  return DS_send_cmd_(p_super, stream);
}


/**
 * @brief  R}hM
 *
 *         DS_send_general_cmd  DS_send_req_tlm_cmd̋ʕ
 * @param  *p_super DriverSuper\̂ւ̃|C^
 * @param  stream   ǂconfiggp邩Dstream0-MAXȂ̂ŁCpENUMȂǐ錾Ďg₷΂ƎvD
 * @retval DS_ERR_CODE_OK  : I
 * @retval DS_ERR_CODE_ERR : IF_TX ł̃G[
 * @note   M󋵂G[ send_status_ Ɋi[Ă
 */
static DS_ERR_CODE DS_send_cmd_(DriverSuper* p_super, uint8_t stream)
{
  DS_StreamConfig* p_stream_config = &(p_super->stream_config[stream]);

  // setter  validation ƁCȂǂŉx邱ƂC
  //  validation ͑łグƂނn㎎ɗLpȂ̂ŁCɒu
  if (p_stream_config->is_validation_needed_for_send_)
  {
    DS_ERR_CODE ret = DS_validate_stream_config_(p_stream_config);
    if (ret != DS_ERR_CODE_OK)
    {
      p_stream_config->send_status_.status_code = DS_STREAM_SEND_STATUS_VALIDATE_ERR;
      return DS_ERR_CODE_ERR;
    }
  }

  p_stream_config->send_status_.ret_from_if_tx = DS_tx_(p_super, stream);

  if (p_stream_config->send_status_.ret_from_if_tx != 0)
  {
    p_stream_config->send_status_.status_code = DS_STREAM_SEND_STATUS_TX_ERR;
    return DS_ERR_CODE_ERR;
  }

  p_stream_config->send_status_.status_code = DS_STREAM_SEND_STATUS_OK;
  return DS_ERR_CODE_OK;
}

/**
 * @brief  p̋@ɃR}h𔭍s
 * @note   ̊֐̎sOɁCtx_frame_, tx_frame_size_̐ݒ肪Kvł
 * @param  *p_super DriverSuper\̂ւ̃|C^
 * @param  stream   ǂconfiggp邩Dstream0-MAXȂ̂ŁCpENUMȂǐ錾Ďg₷΂ƎvD
 * @retval DS_ERR_CODE_OK (0) : I
 * @retval 0ȊO : IF_TX ̖߂l
 */
static int DS_tx_(DriverSuper* p_super, uint8_t stream)
{
  int ret;
  DS_StreamConfig* p_stream_config = &(p_super->stream_config[stream]);

  if (p_stream_config->tx_frame_size_ == 0) return DS_ERR_CODE_OK;
  if (p_stream_config->tx_frame_ == NULL) return DS_ERR_CODE_OK;

#ifdef DS_DEBUG
  Printf("DS: tx_\n");
#endif

  ret = (*IF_TX[p_super->interface])(p_super->if_config,
                                     p_stream_config->tx_frame_,
                                     (int)p_stream_config->tx_frame_size_);

  if (ret != 0) return ret;
  return DS_ERR_CODE_OK;
}


/**
 * @brief  p̋@킩̎Mf[^邩mFCM
 * @param  *p_super DriverSuper\̂ւ̃|C^
 * @retval 0    : Mf[^Ȃ
 * @retval  : Mf[^ [Byte]
 * @retval  : IF_RX̃G[
 */
static int DS_rx_(DriverSuper* p_super)
{
  int flag;
  int rec_data_len;
  uint8_t stream;
#ifdef DS_DEBUG_SHOW_REC_DATA
  int16_t i;
#endif

  // ȂƂPstreamLłCrx_frame_size_0ȊOłȂƎM͂Ȃ
  flag = 0;
  for (stream = 0; stream < DS_STREAM_MAX; ++stream)
  {
    if (! (p_super->stream_config[stream].is_enabled_) ) continue;
    if (p_super->stream_config[stream].rx_frame_size_ != 0)
    {
      flag = 1;
    }
  }
  if (flag == 0) return 0;

  rec_data_len = (*IF_RX[p_super->interface])(p_super->if_config,
                                              p_super->config.rx_buffer_,
                                              DS_RX_BUFFER_SIZE_MAX);

#ifdef DS_DEBUG
  Printf("DS: rx_\n");
#endif

  if (rec_data_len <= 0) return rec_data_len;     // ̓G[R[h

#ifdef DS_DEBUG_SHOW_REC_DATA
  Printf("DS: Receive data size is %d bytes, as follow:\n", rec_data_len);
  for (i = 0; i < rec_data_len; i++)
  {
    Printf("%02x ", p_super->config.rx_buffer_[i]);
    if (i % 4 == 3) Printf("   ");
  }
  Printf("\n");
#endif

  return rec_data_len;
}


/**
 * @brief  Mt[͊֐
 * @param  *p_super      DriverSuper\̂ւ̃|C^
 * @param  stream        ǂconfiggp邩Dstream0-MAXȂ̂ŁCpENUMȂǐ錾Ďg₷΂ƎvD
 * @param  rec_data_len  Mf[^̃obt@̒
 * @return void ڍׂ DS_StreamRecStatus
 */
static void DS_analyze_rx_buffer_(DriverSuper* p_super,
                                  uint8_t stream,
                                  uint16_t rec_data_len)
{
  // ͗pMobt@
  // ȃf[^Ȃ̂ŁCstaticŗ\ߊmۂĂCs̃X^bN͊
  static uint8_t rx_buffer[DS_RX_BUFFER_SIZE_MAX * 2];
  DS_StreamConfig* p_stream_config = &(p_super->stream_config[stream]);
  uint16_t total_processed_data_len;

  rec_data_len = DS_analyze_rx_buffer_prepare_buffer_(p_super, stream, rx_buffer, rec_data_len);

  total_processed_data_len = DS_analyze_rx_buffer_pickup_(p_stream_config, rx_buffer, rec_data_len);

  DS_analyze_rx_buffer_carry_over_buffer_(p_stream_config, rx_buffer, total_processed_data_len, rec_data_len);
}


/**
 * @brief  ͗pMobt@̏
 *
 *         Jzꂽf[^ƍMf[^̌sCMf[^͂̏
 * @param[in]  *p_super      DriverSuper\̂ւ̃|C^
 * @param[in]  stream        ǂconfiggp邩Dstream0-MAXȂ̂ŁCpENUMȂǐ錾Ďg₷΂ƎvD
 * @param[out] *rx_buffer    ͗pMobt@
 * @param[in]  rec_data_len  Mf[^̃obt@̒
 * @return ͗pMobt@̒
 */
static uint16_t DS_analyze_rx_buffer_prepare_buffer_(DriverSuper* p_super,
                                                     uint8_t stream,
                                                     uint8_t* rx_buffer,
                                                     uint16_t rec_data_len)
{
  DS_StreamConfig* p_stream_config = &(p_super->stream_config[stream]);
  uint16_t buffer_offset = 0;

  // JzMf[^̂Ƃ肱
  if (p_stream_config->is_rx_buffer_carry_over_)
  {
    memcpy(rx_buffer,
           p_stream_config->rx_buffer_for_carry_over_,
           (size_t)p_stream_config->carry_over_buffer_size_);
    buffer_offset += p_stream_config->carry_over_buffer_size_;
  }

  // M̂Ƃ肱
  memcpy(&(rx_buffer[buffer_offset]),
         p_super->config.rx_buffer_,
         (size_t)rec_data_len);

  if (p_stream_config->is_rx_buffer_carry_over_)
  {
    rec_data_len += p_stream_config->carry_over_buffer_size_;
  }

  return rec_data_len;
}


/**
 * @brief  t[͊֐
 *
 *         ͗pMobt@𑖍CoCgPʂŃt[f[^EĂ
 * @param  p_stream_config DriverSuper\̂DS_StreamConfig
 * @param  rx_buffer       Mf[^̃obt@izj
 * @param  rec_data_len    Mf[^̃obt@̒
 * @return ̌ĂяoőoCgD̑̏ڍׂ DS_StreamRecStatus
 */
static uint16_t DS_analyze_rx_buffer_pickup_(DS_StreamConfig* p_stream_config,
                                             uint8_t* rx_buffer,
                                             uint16_t rec_data_len)
{
  uint16_t total_processed_data_len;
  // Mobt@̃f[^𑖍CKvȃf[^t[Ƃpickup֐
  uint16_t (*pickup_func)(DS_StreamConfig* p_stream_config,
                          uint8_t* rx_buffer,
                          uint16_t total_processed_data_len,
                          uint16_t rec_data_len);
  p_stream_config->rx_frame_head_pos_of_frame_candidate_ = 0;

  // TODO: rbOf[^Ή
  if (p_stream_config->rx_frame_size_ > 0 && p_stream_config->rx_frame_size_ < DS_RX_FRAME_SIZE_MAX)
  {
    pickup_func = DS_analyze_rx_buffer_fixed_pickup_;
  }
  else if (p_stream_config->rx_frame_size_ > 0)
  {
    // DS_analyze_rx_buffer_fixed_bigdata_(void)
    pickup_func = NULL;
    return rec_data_len;
  }
  else if (p_stream_config->rx_frame_size_ < 0 && p_stream_config->rx_frame_size_ < DS_RX_FRAME_SIZE_MAX)
  {
    // t[Ƀt[f[^܂܂Ă邩H
    if (p_stream_config->rx_framelength_pos_ >= 0)
    {
      pickup_func = DS_analyze_rx_buffer_variable_pickup_with_rx_frame_size_;
    }
    else
    {
      pickup_func = DS_analyze_rx_buffer_variable_pickup_with_footer_;
    }
  }
  else
  {
    // DS_analyze_rx_buffer_variable_bigdata_(void)
    pickup_func = NULL;
    return rec_data_len;
  }

  // Mobt@f[^sbNAbvĂ
  total_processed_data_len = p_stream_config->carry_over_buffer_next_pos_;
  while (total_processed_data_len < rec_data_len)
  {
    uint16_t processed_data_len = pickup_func(p_stream_config,
                                              rx_buffer,
                                              total_processed_data_len,
                                              rec_data_len);
    total_processed_data_len += processed_data_len;

    if (p_stream_config->rec_status_.status_code == DS_STREAM_REC_STATUS_FIXED_FRAME)
    {
      break;
    }

    // sNC݂frame̐擪 + 1oCgڂɑꏊ߂
    if (p_stream_config->rec_status_.status_code == DS_STREAM_REC_STATUS_HEADER_MISMATCH ||
        p_stream_config->rec_status_.status_code == DS_STREAM_REC_STATUS_FOOTER_MISMATCH ||
        p_stream_config->rec_status_.status_code == DS_STREAM_REC_STATUS_RX_FRAME_TOO_LONG ||
        p_stream_config->rec_status_.status_code == DS_STREAM_REC_STATUS_RX_FRAME_TOO_SHORT)
    {
      total_processed_data_len = (uint16_t)(p_stream_config->rx_frame_head_pos_of_frame_candidate_ + 1);

      // ̕ł̏̂߂ɁCڍ׃G[݂̃Xe[^Xɏ㏑
      p_stream_config->rec_status_.status_code = DS_STREAM_REC_STATUS_FINDING_HEADER;
    }

    // processed_data_len > rec_data_lenȂ邱Ƃ͂肦ȂCÖ׃`bNHH
  }

  return total_processed_data_len;
}


/**
 * @brief  t[͊֐̃f[^Jz֐
 * @param  p_stream_config          DriverSuper\̂DS_StreamConfig
 * @param  rx_buffer                Mf[^̃obt@izj
 * @param  total_processed_data_len Mf[^̃obt@̂CłɏꂽoCg
 * @param  rec_data_len             Mf[^̃obt@̒
 * @return void
 */
static void DS_analyze_rx_buffer_carry_over_buffer_(DS_StreamConfig* p_stream_config,
                                                    uint8_t* rx_buffer,
                                                    uint16_t total_processed_data_len,
                                                    uint16_t rec_data_len)
{
  // Mf[^ĂяoɈpi̃p^[j
  // t[m肵ꍇi͎Mf[^\j
  //     iȃt[T
  //          --> t[m肵f[^OKƂCȌ̉͂łȂMf[^݂̂Ɉp
  //     iȃt[TL
  //          --> mt[[U[Œe\lCmt[̐擪 + 1 oCgڈȍ~Ɉp
  //              ̃wb_T͂n܂
  // t[m肵Ȃꍇi͎Mf[^͂Ȃj
  //     DS_STREAM_REC_STATUS_FINDING_HEADER ̂Ƃ
  //          --> wb_ȂƂƂȂ̂ŁCpf[^͂Ȃ
  //     DS_STREAM_REC_STATUS_FINDING_HEADER łȂƂ
  //          --> t[̃t[擪ȍ~Ɉp
  //              ɂCtb^svȂǂ̕sꍇɃt[T蒼

  p_stream_config->carry_over_buffer_size_ = 0;
  if (p_stream_config->rec_status_.status_code == DS_STREAM_REC_STATUS_FIXED_FRAME)
  {
    if (p_stream_config->is_strict_frame_search_)
    {
      // mt[̐擪 + 1 oCgڈȍ~Ɉp
      p_stream_config->carry_over_buffer_size_ = (uint16_t)(rec_data_len - p_stream_config->rx_frame_head_pos_of_frame_candidate_ - 1);
      // ́Cpf[^̐擪Ăуt[T
      p_stream_config->carry_over_buffer_next_pos_ = 0;
    }
    else
    {
      // t[m肵āC͂łȂMf[^ꍇCɈp
      p_stream_config->carry_over_buffer_size_ = (uint16_t)(rec_data_len - total_processed_data_len);
      // ́Cpf[^̐擪Ăуt[T
      p_stream_config->carry_over_buffer_next_pos_ = 0;
    }
  }
  else
  {
    if (p_stream_config->rec_status_.status_code = DS_STREAM_REC_STATUS_FINDING_HEADER)
    {
      // pf[^͂Ȃ
      p_stream_config->carry_over_buffer_size_ = 0;
      // ́C擪Ăуt[T
      p_stream_config->carry_over_buffer_next_pos_ = 0;
    }
    else
    {
      // mt[̐擪ȍ~Ɉp
      p_stream_config->carry_over_buffer_size_ = (uint16_t)(rec_data_len - p_stream_config->rx_frame_head_pos_of_frame_candidate_);
      // ́Cpf[^̓XLbvCM̂̐擪t[
      p_stream_config->carry_over_buffer_next_pos_ = p_stream_config->carry_over_buffer_size_;
    }
  }

  if (p_stream_config->carry_over_buffer_size_ > 0 && p_stream_config->carry_over_buffer_size_ <= DS_RX_BUFFER_SIZE_MAX)
  {
    uint16_t pos = (uint16_t)(rec_data_len - p_stream_config->carry_over_buffer_size_);
    p_stream_config->is_rx_buffer_carry_over_ = 1;
    memcpy(p_stream_config->rx_buffer_for_carry_over_,
           &(rx_buffer[pos]),
           (size_t)p_stream_config->carry_over_buffer_size_);
  }
  else
  {
    // pTCY DS_RX_BUFFER_SIZE_MAX 𒴂ꍇC̃Lp𒴂Ă܂Ă̂ŁCZbgD
    if (p_stream_config->carry_over_buffer_size_ > 0)
    {
      p_stream_config->rec_status_.count_of_carry_over_failures++;
    }
    p_stream_config->is_rx_buffer_carry_over_    = 0;
    p_stream_config->carry_over_buffer_size_     = 0;
    p_stream_config->carry_over_buffer_next_pos_ = 0;
  }

  return;
}


/**
 * @brief  Œ蒷t[͊֐ioCg񏈗j
 *
 *         Mobt@̃f[^𑖍CKvȃf[^t[Ƃpickup
 * @param  p_stream_config          DriverSuper\̂DS_StreamConfig
 * @param  rx_buffer                Mf[^̃obt@izj
 * @param  total_processed_data_len Mf[^̃obt@̂CłɏꂽoCg
 * @param  rec_data_len             Mf[^̃obt@̒
 * @return ̌ĂяoőoCgD̑̏ڍׂ DS_StreamRecStatus
 */
static uint16_t DS_analyze_rx_buffer_fixed_pickup_(DS_StreamConfig* p_stream_config,
                                                   uint8_t* rx_buffer,
                                                   uint16_t total_processed_data_len,
                                                   uint16_t rec_data_len)
{
  uint16_t unprocessed_data_len = (uint16_t)(rec_data_len - total_processed_data_len);      // ̃LXg͎኱ȂiR[h_IɐΖȂj
  DS_StreamConfig* p = p_stream_config;  // ƕϐĔzindexȂǂ݂炢̂...

  if (p->rx_frame_rec_len_ == 0 && p->rx_header_size_ != 0)
  {
    // ܂wb_̐擪疢̏ꍇiwb_Ȃ͂̓XLbvj
    return DS_analyze_rx_buffer_finding_header_(p_stream_config,
                                                rx_buffer,
                                                total_processed_data_len,
                                                rec_data_len);
  }
  else if (p->rx_frame_rec_len_ < p->rx_header_size_)
  {
    // wb_M
    return DS_analyze_rx_buffer_receiving_header_(p_stream_config,
                                                  rx_buffer,
                                                  total_processed_data_len);
  }
  else if (p->rx_frame_rec_len_ < p->rx_frame_size_ - p->rx_footer_size_)
  {
    // f[^M
    // ͍̂߂Ɉꊇ
    uint16_t pickup_data_len;

    // wb_Ȃ̏ꍇ́Ct[擪
    if (p->rx_frame_rec_len_ == 0)
    {
      p->rx_frame_head_pos_of_frame_candidate_ = total_processed_data_len;
    }

    pickup_data_len = (uint16_t)(p->rx_frame_size_ - p->rx_footer_size_ - p->rx_frame_rec_len_);

    // őSMȂꍇ
    if (pickup_data_len > unprocessed_data_len)
    {
      pickup_data_len = unprocessed_data_len;
    }

    memcpy(&(p->rx_frame_[p->rx_frame_rec_len_]),
           &(rx_buffer[total_processed_data_len]),
           (size_t)pickup_data_len);

    p->rx_frame_rec_len_ += pickup_data_len;
    p->rec_status_.status_code = DS_STREAM_REC_STATUS_RECEIVING_DATA;

    // tb^ȂCdataMd؂ꍇ̓t[m
    // ꂪȂƁCDS_analyze_rx_buffer_fixed_ 
    // ܂ɎMf[^҂t[ꍇɁCt[m肪Px邱ƂɂȂ̂
    if (p->rx_footer_size_ == 0 && p->rx_frame_rec_len_ == p->rx_frame_size_)
    {
      p->rec_status_.status_code = DS_STREAM_REC_STATUS_FIXED_FRAME;
      p->rec_status_.fixed_frame_len = p->rx_frame_rec_len_;
      p->rx_frame_rec_len_ = 0;
    }

    return pickup_data_len;
  }
  else
  {
    // tb^M or tb^Ȃ̏ꍇ̓t[m
    return DS_analyze_rx_buffer_receiving_footer_(p_stream_config,
                                                  rx_buffer,
                                                  total_processed_data_len,
                                                  (uint16_t)(p->rx_frame_size_));
  }
}


/**
 * @brief  σt[͊֐ioCg񏈗j
 *
 *         Mobt@̃f[^𑖍CKvȃf[^t[Ƃpickup
 * @note   Mt[Ƀt[f[^݂Ă邱ƂOƂ
 * @param  p_stream_config          DriverSuper\̂DS_StreamConfig
 * @param  rx_buffer                Mf[^̃obt@izj
 * @param  total_processed_data_len Mf[^̃obt@̂CłɏꂽoCg
 * @param  rec_data_len             Mf[^̃obt@̒
 * @return ̌ĂяoőoCgD̑̏ڍׂ DS_StreamRecStatus
 */
static uint16_t DS_analyze_rx_buffer_variable_pickup_with_rx_frame_size_(DS_StreamConfig* p_stream_config,
                                                                         uint8_t* rx_buffer,
                                                                         uint16_t total_processed_data_len,
                                                                         uint16_t rec_data_len)
{
  uint16_t unprocessed_data_len = (uint16_t)(rec_data_len - total_processed_data_len);      // ̃LXg͎኱ȂiR[h_IɐΖȂj
  DS_StreamConfig* p = p_stream_config;  // ƕϐĔzindexȂǂ݂炢̂...
  uint32_t rx_frame_size = DS_analyze_rx_buffer_get_framelength_(p_stream_config);      // ܂MĂȂꍇ͕sl邱Ƃɗӂ邱ƁII

  if (p->rx_frame_rec_len_ == 0 && p->rx_header_size_ != 0)
  {
    // ܂wb_̐擪疢̏ꍇiwb_Ȃ͂̓XLbvj
    return DS_analyze_rx_buffer_finding_header_(p_stream_config,
                                                rx_buffer,
                                                total_processed_data_len,
                                                rec_data_len);
  }
  else if (p->rx_frame_rec_len_ < p->rx_header_size_)
  {
    // wb_M
    return DS_analyze_rx_buffer_receiving_header_(p_stream_config,
                                                  rx_buffer,
                                                  total_processed_data_len);
  }
  else if (p->rx_frame_rec_len_ < p->rx_framelength_pos_ + p->rx_framelength_type_size_)
  {
    // t[TCYT
    // ͍̂߂Ɉꊇ
    uint16_t pickup_data_len;

    // wb_Ȃ̏ꍇ́Ct[擪
    if (p->rx_frame_rec_len_ == 0)
    {
      p->rx_frame_head_pos_of_frame_candidate_ = total_processed_data_len;
    }

    pickup_data_len = (uint16_t)(p->rx_framelength_pos_ + p->rx_framelength_type_size_ - p->rx_frame_rec_len_);

    // őSMȂꍇ
    if (pickup_data_len > unprocessed_data_len)
    {
      pickup_data_len = unprocessed_data_len;
    }

    memcpy(&(p->rx_frame_[p->rx_frame_rec_len_]),
           &(rx_buffer[total_processed_data_len]),
           (size_t)pickup_data_len);

    p->rx_frame_rec_len_ += pickup_data_len;
    p->rec_status_.status_code = DS_STREAM_REC_STATUS_RECEIVING_FRAMELENGTH;

    // t[MIꍇC`FbN
    if (p->rx_frame_rec_len_ >= p->rx_framelength_pos_ + p->rx_framelength_type_size_)
    {
      rx_frame_size = DS_analyze_rx_buffer_get_framelength_(p_stream_config);

      // obt@[̓G[oI
      if (rx_frame_size > DS_RX_FRAME_SIZE_MAX)
      {
        p->rec_status_.status_code = DS_STREAM_REC_STATUS_RX_FRAME_TOO_LONG;
        p->rx_frame_rec_len_ = 0;
#ifdef DS_DEBUG
        Printf("DS: RX frame size is too long\n");
#endif
        return 0;       // TODO: ŉԂ͂lĂǂH
      }

      // bodyTCY0ȏv
      if (rx_frame_size < p->rx_header_size_ + p->rx_footer_size_)
      {
        p->rec_status_.status_code = DS_STREAM_REC_STATUS_RX_FRAME_TOO_SHORT;
        p->rx_frame_rec_len_ = 0;
#ifdef DS_DEBUG
        Printf("DS: RX frame size is too short\n");
#endif
        return 0;       // TODO: ŉԂ͂lĂǂH
      }
    }

    return pickup_data_len;
  }
  else if (p->rx_frame_rec_len_ < rx_frame_size - p->rx_footer_size_)
  {
    // f[^M
    // ͍̂߂Ɉꊇ

    uint16_t pickup_data_len = (uint16_t)(rx_frame_size - p->rx_footer_size_ - p->rx_frame_rec_len_);    // TODO: ݁Ct[uint16_t𒴂邱Ƃ͑z肵ĂȂI

    // őSMȂꍇ
    if (pickup_data_len > unprocessed_data_len)
    {
      pickup_data_len = unprocessed_data_len;
    }

    memcpy(&(p->rx_frame_[p->rx_frame_rec_len_]),
           &(rx_buffer[total_processed_data_len]),
           (size_t)pickup_data_len);

    p->rx_frame_rec_len_ += pickup_data_len;
    p->rec_status_.status_code = DS_STREAM_REC_STATUS_RECEIVING_DATA;

    // tb^ȂCdataMd؂ꍇ̓t[m
    // ꂪȂƁCDS_analyze_rx_buffer_fixed_ ō܂ɎMf[^҂t[ꍇɁCt[m肪Px邱ƂɂȂ̂
    if (p->rx_footer_size_ == 0 && p->rx_frame_rec_len_ == rx_frame_size)
    {
      p->rec_status_.status_code = DS_STREAM_REC_STATUS_FIXED_FRAME;
      p->rx_frame_rec_len_ = 0;
    }

    return pickup_data_len;
  }
  else
  {
    // tb^M or tb^Ȃ̏ꍇ̓t[m
    return DS_analyze_rx_buffer_receiving_footer_(p_stream_config,
                                                  rx_buffer,
                                                  total_processed_data_len,
                                                  (uint16_t)rx_frame_size);
  }
}


/**
 * @brief  σt[͊֐ioCg񏈗j
 *
 *         Mobt@̃f[^𑖍CKvȃf[^t[Ƃpickup
 * @note   DS_analyze_rx_buffer_variable_pickup_with_rx_frame_size_ Ƃ̈Ⴂ́Cef[^t[Ɋ܂܂邩ۂ
 * @note   tb^݂Ă邱ƂOƂ
 * @note   wb_Ȃ͔F߂DCMf[^擪t[Ƃ݂Ȃ̂ŁCwb_
 * @param  p_stream_config          DriverSuper\̂DS_StreamConfig
 * @param  rx_buffer                Mf[^̃obt@izj
 * @param  total_processed_data_len Mf[^̃obt@̂CłɏꂽoCg
 * @param  rec_data_len             Mf[^̃obt@̒
 * @return ̌ĂяoőoCgD̑̏ڍׂ DS_StreamRecStatus
 */
static uint16_t DS_analyze_rx_buffer_variable_pickup_with_footer_(DS_StreamConfig* p_stream_config,
                                                                  uint8_t* rx_buffer,
                                                                  uint16_t total_processed_data_len,
                                                                  uint16_t rec_data_len)
{
  uint16_t unprocessed_data_len = (uint16_t)(rec_data_len - total_processed_data_len);      // ̃LXg͎኱ȂiR[h_IɐΖȂj
  DS_StreamConfig* p = p_stream_config;  // ƕϐĔzindexȂǂ݂炢̂...

  if (p->rx_frame_rec_len_ == 0 && p->rx_header_size_ != 0)
  {
    // ܂wb_̐擪疢̏ꍇiwb_Ȃ͂̓XLbvj
    return DS_analyze_rx_buffer_finding_header_(p_stream_config,
                                                rx_buffer,
                                                total_processed_data_len,
                                                rec_data_len);
  }
  else if (p->rx_frame_rec_len_ < p->rx_header_size_)
  {
    // wb_M
    return DS_analyze_rx_buffer_receiving_header_(p_stream_config,
                                                  rx_buffer,
                                                  total_processed_data_len);
  }
  else
  {
    // Ō܂ŎMCtb^̍ŏITDtb^Ȃ͂肦ȂD
    uint8_t* p_footer_last;     // inclusive
    int32_t  body_data_len;     // TCYIɂu16ł悢CƂ肽̂i32ƂĂ
    uint16_t processed_data_len;
    uint16_t i;
    uint16_t estimated_rx_frame_size;
    uint16_t pickup_data_len;

    // wb_Ȃ̏ꍇ́Ct[擪
    if (p->rx_frame_rec_len_ == 0)
    {
      p->rx_frame_head_pos_of_frame_candidate_ = total_processed_data_len;
    }

    // ͂Ăf[^Mt[obt@Ɋi[
    // ͍̂߂Ɉꊇ
    pickup_data_len = unprocessed_data_len;
    // iɃtb^MȂꍇɃobt@[I[o[邱Ƃh
    if (p->rx_frame_rec_len_ + pickup_data_len > DS_RX_FRAME_SIZE_MAX)
    {
      if (p->rx_frame_rec_len_ >= DS_RX_FRAME_SIZE_MAX)
      {
        // ȏMłȂ߁Ctb^TsƂāCZbg
        p->rec_status_.status_code = DS_STREAM_REC_STATUS_RX_FRAME_TOO_LONG;
        p->rx_frame_rec_len_ = 0;
#ifdef DS_DEBUG
        Printf("DS: RX frame is too long\n");
#endif
        return 0;   // ς݃f[^Ȃ
      }
      pickup_data_len = (uint16_t)(DS_RX_FRAME_SIZE_MAX - p->rx_frame_rec_len_);
    }
    memcpy(&(p->rx_frame_[p->rx_frame_rec_len_]),
           &(rx_buffer[total_processed_data_len]),
           (size_t)pickup_data_len);

    // tb^ŏIT
    p_footer_last = (uint8_t*)memchr(&(rx_buffer[p->rx_frame_rec_len_]),
                                     (int)(p->rx_footer_[p->rx_footer_size_ - 1]),
                                     (size_t)pickup_data_len);

    if (p_footer_last == NULL)
    {
      // ܂܂M
      p->rx_frame_rec_len_ += pickup_data_len;
      p->rec_status_.status_code = DS_STREAM_REC_STATUS_RECEIVING_DATA;
      return pickup_data_len;
    }

    processed_data_len = (uint16_t)(p_footer_last - &(rx_buffer[p->rx_frame_rec_len_]) + 1);
    body_data_len = (p_footer_last - rx_buffer + 1) - p->rx_header_size_ - p->rx_footer_size_;
    if (body_data_len < 0)
    {
      // ̓tb^ł͂Ȃ̂ŎMs
      // ܂܂M
      p->rx_frame_rec_len_ += pickup_data_len;
      p->rec_status_.status_code = DS_STREAM_REC_STATUS_RECEIVING_DATA;
      return pickup_data_len;
    }

    // tb^┭
    // tb^`FbN
    estimated_rx_frame_size = (uint16_t)(p->rx_header_size_ + body_data_len + p->rx_footer_size_);
    for (i = 0; i < p->rx_footer_size_; i++)
    {
      if (rx_buffer[estimated_rx_frame_size - i - 1] != p->rx_footer_[p->rx_footer_size_ - i - 1])
      {
        // ̓tb^ł͂Ȃ̂ŎMs
        // ܂܂M
        p->rx_frame_rec_len_ += pickup_data_len;
        p->rec_status_.status_code = DS_STREAM_REC_STATUS_RECEIVING_DATA;
        return pickup_data_len;
      }
    }

    // tb^m  t[m
    p->rec_status_.status_code = DS_STREAM_REC_STATUS_FIXED_FRAME;
    p->rec_status_.fixed_frame_len = estimated_rx_frame_size;
    p->rx_frame_rec_len_ = 0;
    return processed_data_len;
  }
}


/**
 * @brief  t[͊֐iwb_Tj
 * @note   wb_ꍇCŏ1 bytêݏ
 * @param  p_stream_config          DriverSuper\̂DS_StreamConfig
 * @param  rx_buffer                Mf[^̃obt@izj
 * @param  total_processed_data_len Mf[^̃obt@̂CłɏꂽoCg
 * @param  rec_data_len             Mf[^̃obt@̒
 * @return ̌ĂяoőoCgD̑̏ڍׂ DS_StreamRecStatus
 */
static uint16_t DS_analyze_rx_buffer_finding_header_(DS_StreamConfig* p_stream_config,
                                                     uint8_t* rx_buffer,
                                                     uint16_t total_processed_data_len,
                                                     uint16_t rec_data_len)
{
  uint16_t unprocessed_data_len = (uint16_t)(rec_data_len - total_processed_data_len);      // ̃LXg͎኱ȂiR[h_IɐΖȂj
  DS_StreamConfig* p = p_stream_config;  // ƕϐĔzindexȂǂ݂炢̂...
  uint8_t* p_header;
  uint16_t processed_data_len;

  // R[h΁Cwb_ȂP[X͂ɓBȂCk|Ă
  if (p_stream_config->rx_header_ == NULL)
  {
#ifdef DS_DEBUG
    Printf("DS: RX header is invalid\n");
#endif
    p->rec_status_.status_code = DS_STREAM_REC_STATUS_OTHER_ERR;
    return unprocessed_data_len;
  }

  // ܂wb_̐擪疢̏ꍇ
  // ͍̂߂Ɉꊇ
  p_header = (uint8_t*)memchr(&(rx_buffer[total_processed_data_len]),
                              (int)(p->rx_header_[0]),
                              (size_t)unprocessed_data_len);

  if (p_header == NULL)
  {
#ifdef DS_DEBUG
    Printf("DS: RX header not found\n");
#endif
    p->rec_status_.status_code = DS_STREAM_REC_STATUS_FINDING_HEADER;
    return unprocessed_data_len;
  }

  processed_data_len = (uint16_t)(p_header - &(rx_buffer[total_processed_data_len]) + 1);

  // wb_Rs[DzgbufferRsׂǁCƃAhXĂĕ|̂ŁDDD
  p->rx_frame_[p->rx_frame_rec_len_] = p->rx_header_[0];
  p->rx_frame_rec_len_++;

  p->rec_status_.status_code = DS_STREAM_REC_STATUS_RECEIVING_HEADER;
  p->rx_frame_head_pos_of_frame_candidate_ = (uint16_t)(total_processed_data_len + processed_data_len - 1);
  return processed_data_len;
}


/**
 * @brief  t[͊֐iwb_Mj
 * @note   1 bytêݏ
 * @param  p_stream_config          DriverSuper\̂DS_StreamConfig
 * @param  rx_buffer                Mf[^̃obt@izj
 * @param  total_processed_data_len Mf[^̃obt@̂CłɏꂽoCg
 * @return ̌ĂяoőoCgi܂C1Ȃ񂾂ǁjD̑̏ڍׂ DS_StreamRecStatus
 */
static uint16_t DS_analyze_rx_buffer_receiving_header_(DS_StreamConfig* p_stream_config,
                                                       uint8_t* rx_buffer,
                                                       uint16_t total_processed_data_len)
{
  DS_StreamConfig* p = p_stream_config;  // ƕϐĔzindexȂǂ݂炢̂...

  // wb_M
  // Mא؂̂ƂȂǂ̏򂪂߂ǂ̂ŁC1byte

  // wb_H
  if (rx_buffer[total_processed_data_len] == p->rx_header_[p->rx_frame_rec_len_])
  {
    p->rx_frame_[p->rx_frame_rec_len_] = p->rx_header_[p->rx_frame_rec_len_];
    p->rx_frame_rec_len_++;

    p->rec_status_.status_code = DS_STREAM_REC_STATUS_RECEIVING_HEADER;
    return 1;
  }
  else
  {
    // wb_Ȃ
    // DS_STREAM_REC_STATUS_HEADER_MISMATCH ɂȂCĂуobt@߂ăwb_Tn߂
    // ̌ DS_STREAM_REC_STATUS_FINDING_HEADER ɖ߂
    p->rec_status_.status_code = DS_STREAM_REC_STATUS_HEADER_MISMATCH;
    p->rx_frame_rec_len_ = 0;
#ifdef DS_DEBUG
    Printf("DS: RX header is mismatch\n");
#endif
    return 1;
  }
}


/**
 * @brief  t[͊֐itb^Mj
 * @note   1 bytêݏ
 * @note   ݁Ct[uint16_t𒴂邱Ƃ͑z肵ĂȂI
 * @param  p_stream_config          DriverSuper\̂DS_StreamConfig
 * @param  rx_buffer                Mf[^̃obt@izj
 * @param  total_processed_data_len Mf[^̃obt@̂CłɏꂽoCg
 * @param  rx_frame_size            t[TCYiϒt[̏ꍇ̂ŁCɎj
 * @return ̌ĂяoőoCgi܂C1Ȃ񂾂ǁjD̑̏ڍׂ DS_StreamRecStatus
 */
static uint16_t DS_analyze_rx_buffer_receiving_footer_(DS_StreamConfig* p_stream_config,
                                                       uint8_t* rx_buffer,
                                                       uint16_t total_processed_data_len,
                                                       uint16_t rx_frame_size)
{
  DS_StreamConfig* p = p_stream_config;  // ƕϐĔzindexȂǂ݂炢̂...
  uint16_t rec_footer_pos;

  // tb^Ȃ̏ꍇ̓t[m
  if (p->rx_footer_size_ == 0)
  {
    // tb^Ȃ
    p->rec_status_.status_code = DS_STREAM_REC_STATUS_FIXED_FRAME;
    p->rec_status_.fixed_frame_len = p->rx_frame_rec_len_;
    p->rx_frame_rec_len_ = 0;
    return 0;   // ς݃f[^Ȃ
  }

  // tb^M
  // 򂪂߂ǂ̂ŁC1byte
  rec_footer_pos = (uint16_t)(p->rx_frame_rec_len_ - (rx_frame_size - p->rx_footer_size_));

  // ҂Ătb^MłH
  // MłȂꍇC DS_STREAM_REC_STATUS_FOOTER_MISMATCH ɂȂCĂуobt@߂ăwb_Tn߂
  // ̌ DS_STREAM_REC_STATUS_FINDING_HEADER ɖ߂
  if (rx_buffer[total_processed_data_len] != p->rx_footer_[rec_footer_pos])
  {
    p->rec_status_.status_code = DS_STREAM_REC_STATUS_FOOTER_MISMATCH;
    p->rx_frame_rec_len_ = 0;
#ifdef DS_DEBUG
    Printf("DS: RX footer is mismatch\n");
#endif
    return 1;
  }

  // ܂ł琳tb^MĂ
  p->rx_frame_[p->rx_frame_rec_len_] = p->rx_footer_[rec_footer_pos];
  p->rx_frame_rec_len_++;

  if (p->rx_frame_rec_len_ == rx_frame_size)
  {
    // t[m
    p->rec_status_.status_code = DS_STREAM_REC_STATUS_FIXED_FRAME;
    p->rec_status_.fixed_frame_len = p->rx_frame_rec_len_;
    p->rx_frame_rec_len_ = 0;
  }
  else
  {
    // t[m肹
    p->rec_status_.status_code = DS_STREAM_REC_STATUS_RECEIVING_FOOTER;
  }

  return 1;
}


/**
 * @brief  t[͒ɎMt[t[擾֐
 * @note   DS_analyze_rx_buffer_variable_pickup_with_rx_frame_size_ Ă΂邱Ƃz
 * @param  p_stream_config          DriverSuper\̂DS_StreamConfig
 * @return t[
 */
static uint32_t DS_analyze_rx_buffer_get_framelength_(DS_StreamConfig* p_stream_config)
{
  uint32_t len = 0;
  uint8_t  i;

  for (i = 0; i < p_stream_config->rx_framelength_type_size_; ++i)
  {
    if (i == 0)
    {
      len = p_stream_config->rx_frame_[p_stream_config->rx_framelength_pos_];
    }
    else
    {
      len <<= 8;
      len |= p_stream_config->rx_frame_[p_stream_config->rx_framelength_pos_ + i];
    }
  }

  len += p_stream_config->rx_framelength_offset_;
  return len;
}


/**
 * @brief  DS_StreamConfig\̂̏
 *
 *         DS_StreamConfig\̂CftHglŖ߂D
 * @param  p_stream_config  DriverSuper\̂DS_StreamConfig
 * @return DS_ERR_CODE
 */
static DS_ERR_CODE DS_reset_stream_config_(DS_StreamConfig* p_stream_config)
{
  p_stream_config->is_enabled_ = 0;
  p_stream_config->is_strict_frame_search_ = 0;

  p_stream_config->general_cmd_tx_count_               = 0;
  p_stream_config->req_tlm_cmd_tx_count_               = 0;
  p_stream_config->req_tlm_cmd_tx_count_after_last_tx_ = 0;
  p_stream_config->rx_frame_fix_count_                 = 0;

  p_stream_config->general_cmd_tx_time_ = TMGR_get_master_clock();
  p_stream_config->req_tlm_cmd_tx_time_ = TMGR_get_master_clock();
  p_stream_config->rx_frame_fix_time_   = TMGR_get_master_clock();

  p_stream_config->tx_frame_       = NULL;
  p_stream_config->tx_frame_size_  = 0;
  p_stream_config->rx_header_      = NULL;
  p_stream_config->rx_header_size_ = 0;
  p_stream_config->rx_footer_      = NULL;
  p_stream_config->rx_footer_size_ = 0;
  p_stream_config->rx_frame_size_  = 0;

  p_stream_config->rx_framelength_pos_       = -1;
  p_stream_config->rx_framelength_type_size_ = 0;
  p_stream_config->rx_framelength_offset_    = 0;

  p_stream_config->data_analyzer_ = DS_data_analyzer_dummy_;
  p_stream_config->ret_from_data_analyzer_ = DS_ERR_CODE_OK;

  p_stream_config->should_monitor_for_tlm_disruption_ = 0;
  p_stream_config->time_threshold_for_tlm_disruption_ = 60 * 1000;      // ̒l͂悭l邱

  p_stream_config->is_validation_needed_for_send_ = 0;
  p_stream_config->is_validation_needed_for_rec_  = 0;

  p_stream_config->rx_frame_rec_len_ = 0;
  p_stream_config->rx_frame_head_pos_of_frame_candidate_ = 0;

  memset(p_stream_config->rx_frame_,
         0x00,
         sizeof(p_stream_config->rx_frame_));

  p_stream_config->is_rx_buffer_carry_over_ = 0;
  p_stream_config->carry_over_buffer_size_ = 0;
  p_stream_config->carry_over_buffer_next_pos_ = 0;
  memset(p_stream_config->rx_buffer_for_carry_over_,
         0x00,
         sizeof(p_stream_config->rx_buffer_for_carry_over_));

  // DS_StreamSendStatus ̏
  p_stream_config->send_status_.status_code    = DS_STREAM_SEND_STATUS_DISABLE;
  p_stream_config->send_status_.ret_from_if_tx = 0;

  // DS_StreamRecStatus ̏
  p_stream_config->rec_status_.status_code                  = DS_STREAM_REC_STATUS_DISABLE;
  p_stream_config->rec_status_.fixed_frame_len              = 0;
  p_stream_config->rec_status_.tlm_disruption_status        = DS_STREAM_TLM_DISRUPTION_STATUS_OK;
  p_stream_config->rec_status_.count_of_carry_over_failures = 0;

  return DS_ERR_CODE_OK;
}


static DS_ERR_CODE DS_validate_stream_config_(DS_StreamConfig* p_stream_config)
{
  if (!p_stream_config->is_enabled_) return DS_ERR_CODE_OK;

  if (p_stream_config->tx_frame_size_  != 0 && p_stream_config->tx_frame_  == NULL) return DS_ERR_CODE_ERR;
  if (p_stream_config->rx_header_size_ != 0 && p_stream_config->rx_header_ == NULL) return DS_ERR_CODE_ERR;
  if (p_stream_config->rx_footer_size_ != 0 && p_stream_config->rx_footer_ == NULL) return DS_ERR_CODE_ERR;

  if (p_stream_config->rx_frame_size_ > DS_RX_FRAME_SIZE_MAX) return DS_ERR_CODE_ERR;   // [TODO] ݂BigDataiڍׂ̓wb_t@CQƁĵ߁CŒe

  if (p_stream_config->rx_frame_size_ < 0)
  {
    // egϒ
    if (p_stream_config->rx_framelength_pos_ < 0)
    {
      // t[TCYf[^Ȃꍇ
      // tb^݂̑K{
      if (p_stream_config->rx_footer_size_ == 0) return DS_ERR_CODE_ERR;
    }
    else
    {
      if (p_stream_config->rx_header_size_ == 0) return DS_ERR_CODE_ERR;       // ϒwb_Ȃ͑ΉȂiŒ蒷̂悤ɂĉDڍׂ̓wb_t@CQƁj
      if (p_stream_config->rx_framelength_pos_ < p_stream_config->rx_header_size_) return DS_ERR_CODE_ERR;    // t[TCYwb_i܂ŒljɊ܂܂邱Ƃ͂肦Ȃ̂
      if (!(p_stream_config->rx_framelength_type_size_ == 1 ||
            p_stream_config->rx_framelength_type_size_ == 2 ||
            p_stream_config->rx_framelength_type_size_ == 3 ||
            p_stream_config->rx_framelength_type_size_ == 4 )) return DS_ERR_CODE_ERR;    // ݂uint8 to uint32̂ݑΉ
    }
  }
  else if (p_stream_config->rx_frame_size_ == 0)
  {
    // eȂ
  }
  else
  {
    // egŒ蒷
    if (p_stream_config->rx_frame_size_ < (p_stream_config->rx_header_size_ + p_stream_config->rx_footer_size_)) return DS_ERR_CODE_ERR;
  }

  if (p_stream_config->is_strict_frame_search_)
  {
    // wb_邱ƂO
    if (p_stream_config->rx_header_size_ == 0) return DS_ERR_CODE_ERR;
  }

  p_stream_config->is_validation_needed_for_send_ = 0;
  p_stream_config->is_validation_needed_for_rec_ = 0;
  return DS_ERR_CODE_OK;
}


static DS_ERR_CODE DS_load_init_setting_dummy_(DriverSuper* p_super)
{
  (void)p_super;
  return DS_ERR_CODE_OK;
}

static DS_ERR_CODE DS_data_analyzer_dummy_(DS_StreamConfig* p_stream_config, void* p_driver)
{
  (void)p_stream_config;
  (void)p_driver;
  return DS_ERR_CODE_OK;
}


// ###### DS_Config Getter/Setter ######
const DS_RecStatus* DSC_get_rec_status(const DriverSuper* p_super)
{
  return &p_super->config.rec_status_;
}

const uint32_t DSC_get_rx_count(const DriverSuper* p_super)
{
  return p_super->config.rx_count_;
}

const uint32_t DSC_get_rx_call_count(const DriverSuper* p_super)
{
  return p_super->config.rx_call_count_;
}

const ObcTime* DSC_get_rx_time(const DriverSuper* p_super)
{
  return &p_super->config.rx_time_;
}

const uint8_t DSC_get_should_monitor_for_rx_disruption(const DriverSuper* p_super)
{
  return p_super->config.should_monitor_for_rx_disruption_;
}

void DSC_enable_monitor_for_rx_disruption(DriverSuper* p_super)
{
  p_super->config.should_monitor_for_rx_disruption_ = 1;
}

void DSC_disable_monitor_for_rx_disruption(DriverSuper* p_super)
{
  p_super->config.should_monitor_for_rx_disruption_ = 0;
}

const uint32_t DSC_get_time_threshold_for_rx_disruption(const DriverSuper* p_super)
{
  return p_super->config.time_threshold_for_rx_disruption_;
}

void DSC_set_time_threshold_for_rx_disruption(DriverSuper* p_super,
                                              const uint32_t time_threshold_for_rx_disruption)
{
  p_super->config.time_threshold_for_rx_disruption_ = time_threshold_for_rx_disruption;
}

const DS_RX_DISRUPTION_STATUS_CODE DSC_get_rx_disruption_status(const DriverSuper* p_super)
{
  return p_super->config.rec_status_.rx_disruption_status;
}

// ###### DS_StreamConfig Getter/Setter ######
const uint8_t DSSC_get_is_enable(const DS_StreamConfig* p_stream_config)
{
  return p_stream_config->is_enabled_;
}

void DSSC_enable(DS_StreamConfig* p_stream_config)
{
  p_stream_config->is_enabled_ = 1;
  p_stream_config->is_validation_needed_for_send_ = 1;
  p_stream_config->is_validation_needed_for_rec_ = 1;
}

void DSSC_disable(DS_StreamConfig* p_stream_config)
{
  p_stream_config->is_enabled_ = 0;
}

const uint8_t DSSC_get_is_strict_frame_search(const DS_StreamConfig* p_stream_config)
{
  return p_stream_config->is_strict_frame_search_;
}

void DSSC_enable_strict_frame_search(DS_StreamConfig* p_stream_config)
{
  p_stream_config->is_strict_frame_search_ = 1;
  p_stream_config->is_validation_needed_for_rec_ = 1;
}

void DSSC_disable_strict_frame_search(DS_StreamConfig* p_stream_config)
{
  p_stream_config->is_strict_frame_search_ = 0;
}

const DS_StreamSendStatus* DSSC_get_send_status(const DS_StreamConfig* p_stream_config)
{
  return &p_stream_config->send_status_;
}

const DS_StreamRecStatus* DSSC_get_rec_status(const DS_StreamConfig* p_stream_config)
{
  return &p_stream_config->rec_status_;
}

const uint32_t DSSC_get_general_cmd_tx_count(const DS_StreamConfig* p_stream_config)
{
  return p_stream_config->general_cmd_tx_count_;
}

const uint32_t DSSC_get_req_tlm_cmd_tx_count(const DS_StreamConfig* p_stream_config)
{
  return p_stream_config->req_tlm_cmd_tx_count_;
}

const uint32_t DSSC_get_req_tlm_cmd_tx_count_after_last_tx(const DS_StreamConfig* p_stream_config)
{
  return p_stream_config->req_tlm_cmd_tx_count_after_last_tx_;
}

const uint32_t DSSC_get_rx_frame_fix_count(const DS_StreamConfig* p_stream_config)
{
  return p_stream_config->rx_frame_fix_count_;
}

const ObcTime* DSSC_get_general_cmd_tx_time(const DS_StreamConfig* p_stream_config)
{
  return &p_stream_config->general_cmd_tx_time_;
}

const ObcTime* DSSC_get_req_tlm_cmd_tx_time(const DS_StreamConfig* p_stream_config)
{
  return &p_stream_config->req_tlm_cmd_tx_time_;
}

const ObcTime* DSSC_get_rx_frame_fix_time(const DS_StreamConfig* p_stream_config)
{
  return &p_stream_config->rx_frame_fix_time_;
}

void DSSC_set_tx_frame(DS_StreamConfig* p_stream_config,
                       uint8_t* tx_frame)
{
  p_stream_config->tx_frame_ = tx_frame;
  p_stream_config->is_validation_needed_for_send_ = 1;
}

void DSSC_set_tx_frame_size(DS_StreamConfig* p_stream_config,
                            const uint16_t tx_frame_size)
{
  p_stream_config->tx_frame_size_ = tx_frame_size;
  p_stream_config->is_validation_needed_for_send_ = 1;
}

const uint16_t DSSC_get_tx_frame_size(const DS_StreamConfig* p_stream_config)
{
  return p_stream_config->tx_frame_size_;
}

const uint8_t* DSSC_get_rx_frame(const DS_StreamConfig* p_stream_config)
{
  return p_stream_config->rx_frame_;
}

void DSSC_set_rx_header(DS_StreamConfig* p_stream_config,
                        const uint8_t* rx_header,
                        const uint16_t rx_header_size)
{
  p_stream_config->rx_header_ = rx_header;
  p_stream_config->rx_header_size_ = rx_header_size;
  p_stream_config->is_validation_needed_for_rec_ = 1;
}

void DSSC_set_rx_footer(DS_StreamConfig* p_stream_config,
                        const uint8_t* rx_footer,
                        const uint16_t rx_footer_size)
{
  p_stream_config->rx_footer_ = rx_footer;
  p_stream_config->rx_footer_size_ = rx_footer_size;
  p_stream_config->is_validation_needed_for_rec_ = 1;
}

void DSSC_set_rx_frame_size(DS_StreamConfig* p_stream_config,
                            const int16_t rx_frame_size)
{
  p_stream_config->rx_frame_size_ = rx_frame_size;
  p_stream_config->is_validation_needed_for_rec_ = 1;
}

const uint16_t DSSC_get_rx_header_size(const DS_StreamConfig* p_stream_config)
{
  return p_stream_config->rx_header_size_;
}

const uint16_t DSSC_get_rx_footer_size(const DS_StreamConfig* p_stream_config)
{
  return p_stream_config->rx_footer_size_;
}

const int16_t DSSC_get_rx_frame_size(const DS_StreamConfig* p_stream_config)
{
  return p_stream_config->rx_frame_size_;
}

void DSSC_set_rx_framelength_pos(DS_StreamConfig* p_stream_config,
                                 const int16_t rx_framelength_pos)
{
  p_stream_config->rx_framelength_pos_ = rx_framelength_pos;
  p_stream_config->is_validation_needed_for_rec_ = 1;
}

void DSSC_set_rx_framelength_type_size(DS_StreamConfig* p_stream_config,
                                       const uint16_t rx_framelength_type_size)
{
  p_stream_config->rx_framelength_type_size_ = rx_framelength_type_size;
  p_stream_config->is_validation_needed_for_rec_ = 1;
}

void DSSC_set_rx_framelength_offset(DS_StreamConfig* p_stream_config,
                                    const uint16_t rx_framelength_offset)
{
  p_stream_config->rx_framelength_offset_ = rx_framelength_offset;
  p_stream_config->is_validation_needed_for_rec_ = 1;
}

const uint8_t DSSC_get_should_monitor_for_tlm_disruption(const DS_StreamConfig* p_stream_config)
{
  return p_stream_config->should_monitor_for_tlm_disruption_;
}

void DSSC_enable_monitor_for_tlm_disruption(DS_StreamConfig* p_stream_config)
{
  p_stream_config->should_monitor_for_tlm_disruption_ = 1;
  p_stream_config->is_validation_needed_for_rec_ = 1;
}

void DSSC_disable_monitor_for_tlm_disruption(DS_StreamConfig* p_stream_config)
{
  p_stream_config->should_monitor_for_tlm_disruption_ = 0;
  p_stream_config->is_validation_needed_for_rec_ = 1;
}

const uint32_t DSSC_get_time_threshold_for_tlm_disruption(const DS_StreamConfig* p_stream_config)
{
  return p_stream_config->time_threshold_for_tlm_disruption_;
}

void DSSC_set_time_threshold_for_tlm_disruption(DS_StreamConfig* p_stream_config,
                                                const uint32_t time_threshold_for_tlm_disruption)
{
  p_stream_config->time_threshold_for_tlm_disruption_ = time_threshold_for_tlm_disruption;
  p_stream_config->is_validation_needed_for_rec_ = 1;
}

const DS_STREAM_TLM_DISRUPTION_STATUS_CODE DSSC_get_tlm_disruption_status(const DS_StreamConfig* p_stream_config)
{
  return p_stream_config->rec_status_.tlm_disruption_status;
}

void DSSC_set_data_analyzer(DS_StreamConfig* p_stream_config,
                            DS_ERR_CODE (*data_analyzer)(DS_StreamConfig* p_stream_config, void* p_driver))
{
  p_stream_config->data_analyzer_ = data_analyzer;
  p_stream_config->is_validation_needed_for_rec_ = 1;
}

const DS_ERR_CODE DSSC_get_ret_from_data_analyzer(const DS_StreamConfig* p_stream_config)
{
  return p_stream_config->ret_from_data_analyzer_;
}


// ###### DriverėpUtil֐ ######

CCP_EXEC_STS DS_conv_driver_err_to_ccp_exec_sts(DS_DRIVER_ERR_CODE code)
{
  switch (code)
  {
  case DS_DRIVER_ERR_CODE_ILLEGAL_CONTEXT:
  case DS_DRIVER_ERR_CODE_UNKNOWN_ERR:
    // SĂł̂́Cv
    return CCP_EXEC_ILLEGAL_CONTEXT;
  case DS_DRIVER_ERR_CODE_ILLEGAL_PARAMETER:
    return CCP_EXEC_ILLEGAL_PARAMETER;
  case DS_DRIVER_ERR_CODE_ILLEGAL_LENGTH:
    return CCP_EXEC_ILLEGAL_LENGTH;
  default:
    // ɗ͈̂ȉ
    // DS_DRIVER_ERR_CODE_OK
    return CCP_EXEC_SUCCESS;
  }
}


CCP_EXEC_STS DS_conv_cmd_err_to_ccp_exec_sts(DS_CMD_ERR_CODE code)
{
  switch (code)
  {
  case DS_CMD_ILLEGAL_CONTEXT:
    return CCP_EXEC_ILLEGAL_CONTEXT;
  case DS_CMD_ILLEGAL_PARAMETER:
    return CCP_EXEC_ILLEGAL_PARAMETER;
  case DS_CMD_ILLEGAL_LENGTH:
    return CCP_EXEC_ILLEGAL_LENGTH;
  default:
    // ɗ͈̂ȉ̂R
    // DS_CMD_OK
    // DS_CMD_DRIVER_SUPER_ERR
    // DS_CMD_UNKNOWN_ERR
    // Q̃G[Driver̖ŁCŃG[ׂ
    // ł SUCCESSԂ
    return CCP_EXEC_SUCCESS;
  }
}

#pragma section
