#pragma section REPRO
#include "task_dispatcher.h"

#include <stdlib.h>
#include <string.h> // for memcpy

#include "../../CmdTlm/packet_list.h"
#include "../../CmdTlm/block_command_executor.h"
#include "../ModeManager/mode_manager.h"
#include "../TimeManager/time_manager.h"
#include "../AnomalyLogger/anomaly_logger.h"
#include "../EventManager/event_logger.h"
#include <src_user/CmdTlm/block_command_definitions.h>
#include <src_user/CmdTlm/command_definitions.h>
#include "../../Library/print.h"
#include <src_user/Library/VT100.h>

static TDSP_Info TDSP_info_;
const TDSP_Info* const TDSP_info = &TDSP_info_;

// ݂̃TCNŎsׂ^XNꗗۑ
static PacketList task_list_;

/**
 * @brief `TDSP_info_.task_list_id` ɓo^ꂽ BC  TaskListɓWJ.
 * @note  ۂ BC ̓eǂݍ݁AWJĂ̂ `PL_ACK PL_deploy_block_cmd(args)`
 */
static void TDSP_deploy_block_as_task_list_(void);

/**
 * @brief fobO\
 */
static void print_tdsp_status_(void);

void TDSP_initialize(void)
{
  static PL_Node task_stock_[TDSP_TASK_MAX];
  PL_initialize(task_stock_, TDSP_TASK_MAX, &task_list_);

  // ^XNXgAINITIAL[h̃ubNR}hWJ
  TDSP_info_.tskd = CDIS_init(&task_list_);
  TDSP_info_.task_list_id = MM_get_tasklist_id_of_mode(MD_MODEID_START_UP);
  TDSP_deploy_block_as_task_list_();
  TDSP_info_.activated_at = 0;
}

TDSP_ACK TDSP_set_task_list_id(bct_id_t id)
{
  if (id < 0 || id >= BCT_MAX_BLOCKS) return TDSP_INVAILD_BCT_ID;
  if (!BCE_is_active(id)) return TDSP_INACTIVE_BCT_ID;
  if (BCT_get_bc_length(id) == 0) return TDSP_EMPTY_BC;

  TDSP_info_.task_list_id = id;
  return TDSP_SUCCESS;
}

static void TDSP_deploy_block_as_task_list_(void)
{
  PL_ACK ack;

  // {֐̏Master Cycleωꍇoł悤A
  // ܂Master CyclȅXVB
  TDSP_info_.activated_at = TMGR_get_master_total_cycle() + 1;

  ack = PL_deploy_block_cmd(&task_list_, TDSP_info_.task_list_id, 0);

  if (ack != PL_SUCCESS)
  {
#ifndef AL_DISALBE_AT_C2A_CORE
    AL_add_anomaly(AL_CORE_GROUP_TASK_DISPATCHER, TDSP_DEPLOY_FAILED);
#endif
    EL_record_event((EL_GROUP)EL_CORE_GROUP_TASK_DISPATCHER,
                    TDSP_DEPLOY_FAILED,
                    EL_ERROR_LEVEL_HIGH,
                    (uint32_t)ack);
  }
}

void TDSP_execute_pl_as_task_list(void)
{
  // ܂ (1)ubNR}h^XNXgɓWJꂽ̃TCN  (2)݂̃TCN r
  // (1) = (2) ͓̎WJ 1TCN (100ms) ȓȂ̂Ń^XNXgɓo^ꂽR}hs
  // (1) > (2) ͍̎̃TCNŎsׂR}hׂĎsς݂ł
  // (1) < (2) ͓̎WJ 1TCN (100ms) ȏo߂Ă܂Ă̂ŋI

  if (TDSP_info_.activated_at == TMGR_get_master_total_cycle())
  {
    // ^XNXg̐擪R}hs\莞ƌݎr
    PL_ACK ack = PL_check_tl_cmd(&task_list_,
                                 (size_t)(TMGR_get_master_step()));

    switch (ack)
    {
    case PL_TLC_PAST_TIME:
      // s߂Ăꍇ͎sOɃAm}o^B
#ifndef AL_DISALBE_AT_C2A_CORE
      AL_add_anomaly(AL_CORE_GROUP_TASK_DISPATCHER, TDSP_STEP_OVERRUN);
#endif
      EL_record_event((EL_GROUP)EL_CORE_GROUP_TASK_DISPATCHER,
                      TDSP_STEP_OVERRUN,
                      EL_ERROR_LEVEL_LOW,
                      (uint32_t)CCP_get_ti(&PL_get_head(&task_list_)->packet));

      // FALL THROUGH

    case PL_TLC_ON_TIME:
      // s߂ĂA͎ssb^̏ꍇ̓R}hs
      CDIS_dispatch_command(&(TDSP_info_.tskd));

      if (TDSP_info_.tskd.prev.sts != CCP_EXEC_SUCCESS)
      {
        // R}hsɈُ킪ꍇ̓Am}o^B
#ifndef AL_DISALBE_AT_C2A_CORE
        AL_add_anomaly(AL_CORE_GROUP_TASK_DISPATCHER, TDSP_TASK_EXEC_FAILED);
#endif
        EL_record_event((EL_GROUP)EL_CORE_GROUP_TASK_DISPATCHER,
                      TDSP_TASK_EXEC_FAILED,
                      EL_ERROR_LEVEL_HIGH,
                      (uint32_t)(TDSP_info_.tskd.prev.sts + 100));   // CCP_EXEC_STS ܂ނ̂ŁDDDȂƂ
      }

      break;

    case PL_TLC_NOT_YET:
      // casesK؂Aɗ͈̂ȉ̓̏ꍇ
      // E^XNXg
      // ẼR}h̎s܂

      if (PL_count_active_nodes(&task_list_) == 0)
      {
        // task_listȂēx^XNXgWJ
        // sTDSP_info_.activated_atCNĝŁA"̃TCNs҂"ɂȂ
        TDSP_deploy_block_as_task_list_();
      }

      break;

    default:
      // {ɂ͗Ȃ
#ifndef AL_DISALBE_AT_C2A_CORE
      AL_add_anomaly(AL_CORE_GROUP_TASK_DISPATCHER, TDSP_UNKNOWN);
#endif
      EL_record_event((EL_GROUP)EL_CORE_GROUP_TASK_DISPATCHER,
                      TDSP_UNKNOWN,
                      EL_ERROR_LEVEL_HIGH,
                      (uint32_t)ack);
    }
  }
  else if (TDSP_info_.activated_at > TMGR_get_master_total_cycle())
  {
    // ̃TCN̎s҂
    return;
  }
  else if (TDSP_info_.activated_at < TMGR_get_master_total_cycle())
  {
    if ((TDSP_info_.activated_at == 0) && (TMGR_get_master_total_cycle() == OBCT_MAX_CYCLE - 1))
    {
      // TCN̎s҂(TCNI[o[t[O)
      // { TDSP_info_.activated_at > TMGR_get_master_total_cycle() ƂȂ͂A
      // TDSP_info_.activated_atӂ0ɖ߂Ăꍇɗ
      return;
    }
    else
    {
      // 1TCNȓɑSẴR}hsIȂꍇɗ
#ifndef AL_DISALBE_AT_C2A_CORE
      AL_add_anomaly(AL_CORE_GROUP_TASK_DISPATCHER, TDSP_CYCLE_OVERRUN);
#endif
      EL_record_event((EL_GROUP)EL_CORE_GROUP_TASK_DISPATCHER,
                      TDSP_CYCLE_OVERRUN,
                      EL_ERROR_LEVEL_HIGH,
                      0);

      // XgNA->ēWJTCNĎs
      PL_clear_list(&task_list_);
      TDSP_deploy_block_as_task_list_();
    }
  }
}

void TDSP_resync_internal_counter(void)
{
  TDSP_info_.activated_at = TMGR_get_master_total_cycle();
}

CCP_EXEC_STS Cmd_TDSP_SET_TASK_LIST(const CTCP* packet)
{
  // FIXME: u8 ł̂H ܂CCD
  TDSP_ACK ack = TDSP_set_task_list_id((bct_id_t)(CCP_get_param_head(packet)[0]));

  switch (ack)
  {
  case TDSP_SUCCESS:
    return CCP_EXEC_SUCCESS;
  case TDSP_INVAILD_BCT_ID:
    return CCP_EXEC_ILLEGAL_PARAMETER;
  case TDSP_INACTIVE_BCT_ID:
  case TDSP_EMPTY_BC:
  default:
    return CCP_EXEC_ILLEGAL_CONTEXT;
  }
}

AppInfo print_tdsp_status(void)
{
  return create_app_info("tstm", NULL, print_tdsp_status_);
}

void print_tdsp_status_(void)
{
  VT100_erase_line();
  Printf("TASK: BC %d, ERR (TOTAL, STEP, STS) = (%10u, %3u, %d)\n",
         TDSP_info->task_list_id,
         TDSP_info->tskd.prev_err.time.total_cycle,
         TDSP_info->tskd.prev_err.time.step,
         TDSP_info->tskd.prev_err.sts);
}

#pragma section
